﻿using System;
using System.Management;
using System.Collections;
using System.IO.Ports;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Collections.Generic;
using System.Configuration;
using System.Runtime.InteropServices;
using System.IO;
using System.Xml.Linq;
using Zeroconf;

namespace astorino_lib
{
    public class astorino
    {
        private SerialPort COM = new SerialPort();
        private TcpClient Client = new TcpClient();
        private NetworkStream netStream;
        private List<string> programsName = new List<string>();
        private double[,] transPoints = new double[100, 8];
        private double[,] jointPoints = new double[100, 8];
        private double[] Tool1 = new double[6];
        private double[] Tool2 = new double[6];
        private double[] Tool3 = new double[6];
        private string mP = "";
        private string sP = "";
        private string sn = "";
        private int errCode = 0;
        private string frmV;
        private int rS = 0;
        private double[] pose = new double[7] { 0, 0, 0, 0, 0, 0, 0 };
        private double[] abs_vel = new double[7] { 0, 0, 0, 0, 0, 0, 0 };
        private double[] jt_vel = new double[7] { 0, 0, 0, 0, 0, 0, 0 };
        private double[] joints = new double[7] { 0, 0, 0, 0, 0, 0, 0 };
        private double[] home = new double[7] { 0, 0, 0, 0, 0, 0, 0 };
        private double[] forw = new double[7] { 0, 0, 0, 0, 0, 0, 0 };
        private double[] inv = new double[7] { 0, 0, 0, 0, 0, 0, 0 };
        private double[] oat = new double[3] { 0, 0, 0 };
        private double[] euler = new double[3] { 0, 0, 0 };
        private double[] accData = new double[3] { 0, 0, 0 };
        private byte errorCode = 0;
        private byte cpu_temp = 0;
        private const byte ACK = 0x06;
        private const byte NACK = 0xCC;
        private const byte MC = 0xAA;
        private byte[] status = new byte[5];
        private byte[] IO = new byte[18];
        private byte[] _buffer = new byte[4096]; // Buffer to store incoming data
        private int _bufferIndex = 0;
        private byte[] startBytes = new byte[] { 0x01, 0x02 };
        private int _timeOut = 1000;
        private int _motionTimeOut = 300000;
        private string _version = "1.0.0";

        public struct retVal
        {
            public double[] Values { get; private set; }
            public string Name { get; private set; }
            public List<string> Names { get; private set; }
            public int iVal { get; private set; }
            public byte[] bValues { get; private set; }
            public byte ReturnCode { get; private set; }

            public retVal(double[] doubleValues, byte ret) : this()
            {
                Values = doubleValues;
                Name = "";
                Names = new List<string>();
                iVal = -1;
                bValues = new byte[1];
                ReturnCode = ret;
            }

            public retVal(string name, byte ret) : this()
            {
                Values = new double[1];
                Name = name;
                Names = new List<string>();
                iVal = -1;
                bValues = new byte[1];
                ReturnCode = ret;
            }

            public retVal(List<string> names, byte ret) : this()
            {
                Values = new double[1];
                Name = "";
                Names = names;
                iVal = -1;
                bValues = new byte[1];
                ReturnCode = ret;
            }

            public retVal(int val, byte ret) : this()
            {
                Values = new double[1];
                Name = "";
                Names = new List<string>();
                iVal = val;
                bValues = new byte[1];
                ReturnCode = ret;
            }

            public retVal(byte[] array, byte ret) : this()
            {
                Values = new double[1];
                Name = "";
                Names = new List<string>();
                iVal = -1;
                bValues = array;
                ReturnCode = ret;
            }

            public retVal(byte ret) : this()
            {
                Values = new double[1];
                Name = "";
                Names = new List<string>();
                iVal = -1;
                bValues = new byte[1];
                ReturnCode = ret;
            }

        }

        public struct TransformationPoints
        {
            public double[,] points { get; private set; }
            public byte ReturnCode { get; private set; }
            public TransformationPoints(double[,] p, byte ret)
            {
                points = p;
                ReturnCode = ret;
            }

            public TransformationPoints(byte ret)
            {
                points = null;
                ReturnCode = ret;
            }
        }

        public struct JointPoints
        {
            public double[,] points { get; private set; }
            public byte ReturnCode { get; private set; }
            public JointPoints(double[,] p, byte ret)
            {
                points = p;
                ReturnCode = ret;
            }

            public JointPoints(byte ret)
            {
                points = null;
                ReturnCode = ret;
            }
        }

        /* ----------------- Internal Methods ------------------------- */

        private bool sendData(byte[] data)
        {
            if (COM.IsOpen)
            {
                COM.Write(data, 0, data.Length);
                return true;
            }
            if (Client.Connected && Client != null)
            {
                if (netStream != null)
                {
                    try
                    {
                        netStream.Write(data, 0, data.Length);
                        return true;
                    }
                    catch (Exception) { return false; };
                }
            }
            return false;
        }

        private static int ConvertBitsToInt(bool bit0, bool bit1, bool bit2)
        {
            int result = 0;

            if (bit0)
                result |= 1 << 0;      

            if (bit1)
                result |= 1 << 1;    

            if (bit2)
                result |= 1 << 2;

            return result;
        }

        private byte getCRC(byte[] buffer, byte size)
        {
            byte i;
            UInt16 sum = 0;

            for (i = 0; i < size; i++)
                sum += buffer[i];
            byte ret = (byte)sum;
            return (byte)(ret & 0xFF);
        }

        private static string ExtractString(byte[] byteArray, int startIndex)
        {
            int endIndex = Array.IndexOf(byteArray, (byte)0x03, startIndex);
            if (endIndex == -1)
                return "";

            byte[] stringBytes = new byte[endIndex - startIndex];
            Array.Copy(byteArray, startIndex, stringBytes, 0, endIndex - startIndex);
            return Encoding.ASCII.GetString(stringBytes);
        }

        private byte[] ExtractFrame(byte[] startBytes)
        {
            int startIndex = FindStartBytes(startBytes);

            if (startIndex >= 0)
            {
                int frameEndIndex = _bufferIndex; // Assume frame ends at the end of buffer

                // Search for the next occurrence of startBytes to determine the end of the frame
                for (int i = startIndex + startBytes.Length; i <= _bufferIndex - startBytes.Length; i++)
                {
                    if (IsStartBytes(i, startBytes))
                    {
                        frameEndIndex = i;
                        break;
                    }
                }

                int frameLength = frameEndIndex - startIndex;
                byte[] frame = new byte[frameLength];
                Buffer.BlockCopy(_buffer, startIndex, frame, 0, frameLength);

                // Shift remaining data in the buffer
                _bufferIndex -= frameEndIndex;
                Buffer.BlockCopy(_buffer, frameEndIndex, _buffer, 0, _bufferIndex);

                return frame;
            }

            return null;
        }

        private int FindStartBytes(byte[] startBytes)
        {
            for (int i = 0; i <= _bufferIndex - startBytes.Length; i++)
            {
                if (IsStartBytes(i, startBytes))
                    return i;
            }
            return -1;
        }

        private bool IsStartBytes(int index, byte[] startBytes)
        {
            if (index + startBytes.Length > _buffer.Length)
                return false;

            for (int i = 0; i < startBytes.Length; i++)
            {
                if (_buffer[index + i] != startBytes[i])
                    return false;
            }

            return true;
        }

        private byte ValidateResponse(byte[] buffer)
        {
            byte inCRC = buffer[buffer.Length - 1];
            byte calcCRC = getCRC(buffer, (byte)(buffer.Length - 1));
            if (inCRC == calcCRC)
                return 0x00;
            else
            {
                errorCode = 0x01;
                return 0x01;
            }
        }
        private bool IsBitSet(byte b, int pos)
        {
            return (b & (1 << pos)) != 0;
        }

        private int DecodeResponse(byte[] buffer)
        {
            int i = 0;
            switch (buffer[2])
            {
                case ACK:
                    errorCode = 0x00;
                    return 1;
                case NACK:
                    errorCode = buffer[3];
                    return -1;
                case MC:
                    return 2;
                case 0x27:
                    status[0] = buffer[3];
                    status[1] = buffer[4];
                    status[2] = buffer[5];
                    status[3] = buffer[6];
                    status[4] = buffer[7];
                    break;
                case 0x26:
                    for (i = 0; i < 18; i++)
                    {
                        IO[i] = buffer[i + 3];
                    }
                    break;
                case 0x28:

                    for (i = 0; i < 7; i++)
                    {
                        byte[] dataBytes = new byte[4];
                        Array.Copy(buffer, 3 + i * 4, dataBytes, 0, 4);
                        if (BitConverter.IsLittleEndian)
                            Array.Reverse(dataBytes);

                        double temp = (double)BitConverter.ToInt32(dataBytes, 0);
                        joints[i] = temp / 1000.0;
                    }
                    break;
                case 0x29:
                    for (i = 0; i < 7; i++)
                    {
                        byte[] dataBytes = new byte[4];
                        Array.Copy(buffer, 3 + i * 4, dataBytes, 0, 4);
                        if (BitConverter.IsLittleEndian)
                            Array.Reverse(dataBytes);

                        double temp = (double)BitConverter.ToInt32(dataBytes, 0);
                        pose[i] = temp / 1000.0;
                    }
                    break;
                case 0x2A:
                    sn = ExtractString(buffer, 3);
                    break;
                case 0x38:
                    frmV = ExtractString(buffer, 3);
                    break;
                case 0x46:
                    string nm = ExtractString(buffer, 3);
                    if (programsName.Contains(nm) == false)
                        programsName.Add(nm);
                    break;
                case 0x47:
                    mP = ExtractString(buffer, 3);
                    break;
                case 0x57:
                    sP = ExtractString(buffer, 3);
                    break;
                case 0x3D:
                    for (i = 0; i < 7; i++)
                    {
                        byte[] dataBytes = new byte[4];
                        Array.Copy(buffer, 4 + i * 4, dataBytes, 0, 4);
                        if (BitConverter.IsLittleEndian)
                            Array.Reverse(dataBytes);

                        double temp = (double)BitConverter.ToInt32(dataBytes, 0);
                        transPoints[buffer[3], i] = temp / 1000.0;
                    }
                    transPoints[buffer[3], 7] = 1.0;
                    break;
                case 0x3E:
                    for (i = 0; i < 7; i++)
                    {
                        byte[] dataBytes = new byte[4];
                        Array.Copy(buffer, 4 + i * 4, dataBytes, 0, 4);
                        if (BitConverter.IsLittleEndian)
                            Array.Reverse(dataBytes);

                        double temp = (double)BitConverter.ToInt32(dataBytes, 0);
                        jointPoints[buffer[3], i] = temp / 1000.0;
                    }
                    jointPoints[buffer[3], 7] = 1.0;
                    break;
                case 0x56:
                    for (i = 0; i < 6; i++)
                    {
                        byte[] dataBytes = new byte[4];
                        Array.Copy(buffer, 4 + i * 4, dataBytes, 0, 4);
                        if (BitConverter.IsLittleEndian)
                            Array.Reverse(dataBytes);

                        double temp = (double)BitConverter.ToInt32(dataBytes, 0);
                        if (buffer[3] == 1)
                            Tool1[i] = temp / 1000.0;
                        if (buffer[3] == 2)
                            Tool2[i] = temp / 1000.0;
                        if (buffer[3] == 3)
                            Tool3[i] = temp / 1000.0;
                    }
                    break;
                case 0x4A:
                    for (i = 0; i < 7; i++)
                    {
                        byte[] dataBytes = new byte[4];
                        Array.Copy(buffer, 3 + i * 4, dataBytes, 0, 4);
                        if (BitConverter.IsLittleEndian)
                            Array.Reverse(dataBytes);

                        double temp = (double)BitConverter.ToInt32(dataBytes, 0);
                        home[i] = temp / 1000.0;
                    }
                    break;
                case 0x5E:
                    rS = (int)buffer[3];
                    break;
                case 0x54:
                    for (i = 0; i < 7; i++)
                    {
                        byte[] dataBytes = new byte[4];
                        Array.Copy(buffer, 3 + i * 4, dataBytes, 0, 4);
                        if (BitConverter.IsLittleEndian)
                            Array.Reverse(dataBytes);

                        double temp = (double)BitConverter.ToInt32(dataBytes, 0);
                        forw[i] = temp / 1000.0;
                    }
                    break;
                case 0x55:
                    for (i = 0; i < 7; i++)
                    {
                        byte[] dataBytes = new byte[4];
                        Array.Copy(buffer, 3 + i * 4, dataBytes, 0, 4);
                        if (BitConverter.IsLittleEndian)
                            Array.Reverse(dataBytes);

                        double temp = (double)BitConverter.ToInt32(dataBytes, 0);
                        inv[i] = temp / 1000.0;
                    }
                    break;
                case 0x58:
                    for (i = 0; i < 3; i++)
                    {
                        byte[] dataBytes = new byte[4];
                        Array.Copy(buffer, 3 + i * 4, dataBytes, 0, 4);
                        if (BitConverter.IsLittleEndian)
                            Array.Reverse(dataBytes);

                        double temp = (double)BitConverter.ToInt32(dataBytes, 0);
                        oat[i] = temp / 1000.0;
                    }
                    break;
                case 0x59:
                    for (i = 0; i < 3; i++)
                    {
                        byte[] dataBytes = new byte[4];
                        Array.Copy(buffer, 3 + i * 4, dataBytes, 0, 4);
                        if (BitConverter.IsLittleEndian)
                            Array.Reverse(dataBytes);

                        double temp = (double)BitConverter.ToInt32(dataBytes, 0);
                        euler[i] = temp / 1000.0;
                    }
                    break;
                case 0x5B:
                    cpu_temp = buffer[3];
                    break;
                case 0x5F:
                    for (i = 0; i < 7; i++)
                    {
                        byte[] dataBytes = new byte[4];
                        Array.Copy(buffer, 4 + i * 4, dataBytes, 0, 4);
                        if (BitConverter.IsLittleEndian)
                            Array.Reverse(dataBytes);

                        double temp = (double)BitConverter.ToInt32(dataBytes, 0);
                        transPoints[buffer[3], i] = temp / 1000.0;
                    }
                    transPoints[buffer[3], 7] = 1.0;
                    break;
                case 0x60:
                    for (i = 0; i < 7; i++)
                    {
                        byte[] dataBytes = new byte[4];
                        Array.Copy(buffer, 4 + i * 4, dataBytes, 0, 4);
                        if (BitConverter.IsLittleEndian)
                            Array.Reverse(dataBytes);

                        double temp = (double)BitConverter.ToInt32(dataBytes, 0);
                        jointPoints[buffer[3], i] = temp / 1000.0;

                    }
                    jointPoints[buffer[3], 7] = 1.0;
                    break;
                case 0x62:
                    errCode = buffer[3];
                    break;
                case 0x66:
                    for (i = 0; i < 7; i++)
                    {
                        byte[] dataBytes = new byte[4];
                        Array.Copy(buffer, 3 + i * 4, dataBytes, 0, 4);
                        if (BitConverter.IsLittleEndian)
                            Array.Reverse(dataBytes);

                        double temp = (double)BitConverter.ToInt32(dataBytes, 0);
                        abs_vel[i] = temp / 1000.0;
                    }
                    break;
                case 0x67:
                    for (i = 0; i < 7; i++)
                    {
                        byte[] dataBytes = new byte[4];
                        Array.Copy(buffer, 3 + i * 4, dataBytes, 0, 4);
                        if (BitConverter.IsLittleEndian)
                            Array.Reverse(dataBytes);

                        double temp = (double)BitConverter.ToInt32(dataBytes, 0);
                        jt_vel[i] = temp / 1000.0;
                    }
                    break;
                case 0x6E:
                    for (i = 0; i < 3; i++)
                    {
                        byte[] dataBytes = new byte[4];
                        Array.Copy(buffer, 3 + i * 4, dataBytes, 0, 4);
                        if (BitConverter.IsLittleEndian)
                            Array.Reverse(dataBytes);

                        double temp = (double)BitConverter.ToInt32(dataBytes, 0);
                        accData[i] = temp / 1000.0;
                    }
                    break;
                default:
                    return 0;
            }
            return 0;
        }

        private bool SendFrame(string data, byte cmd)
        {
            byte[] startBytes = new byte[] { 0x01, 0x02, cmd };
            byte[] dataBytes = Encoding.ASCII.GetBytes(data);
            byte[] endByte = new byte[] { 0x03 };
            byte[] CRC = new byte[] { 0x00 };

            byte[] frame = new byte[startBytes.Length + dataBytes.Length + endByte.Length + CRC.Length];

            Buffer.BlockCopy(startBytes, 0, frame, 0, startBytes.Length);
            Buffer.BlockCopy(dataBytes, 0, frame, startBytes.Length, dataBytes.Length);
            Buffer.BlockCopy(endByte, 0, frame, startBytes.Length + dataBytes.Length, endByte.Length);

            frame[frame.Length - 1] = getCRC(frame, (byte)(frame.Length - 1));

            return sendData(frame);
        }

        /* ------------------------- Comunication methods --------------------------------*/
        private bool SendData(byte index, byte cmd, double[] values)
        {
            if (values.Length < 6)
            {
                return false;
            }

            byte[] startBytes = new byte[] { 0x01, 0x02, cmd };
            byte[] indexByte = new byte[] { index };
            int len = values.Length;
            byte[] dataBytes = new byte[len * 4]; // 7 double values converted to int32, each int32 is 4 bytes

            for (int i = 0; i < values.Length; i++)
            {
                int intValue = (int)(values[i] * 1000); // Convert to int32 after multiplying by 1000
                byte[] intBytes = BitConverter.GetBytes(intValue);
                Array.Reverse(intBytes); // Ensure bytes are in correct order (big-endian)
                Buffer.BlockCopy(intBytes, 0, dataBytes, i * 4, 4);
            }

            int frameLength = startBytes.Length + indexByte.Length + dataBytes.Length;
            byte[] frame = new byte[frameLength + 1]; // +1 for CRC

            Buffer.BlockCopy(startBytes, 0, frame, 0, startBytes.Length);
            Buffer.BlockCopy(indexByte, 0, frame, startBytes.Length, indexByte.Length);
            Buffer.BlockCopy(dataBytes, 0, frame, startBytes.Length + indexByte.Length, dataBytes.Length);

            // Calculate CRC
            byte crc = getCRC(frame, (byte)frameLength);
            frame[frame.Length - 1] = crc; // Add CRC to the end of the frame

            return sendData(frame);

        }

        private byte[] WaitForResponse(byte[] startBytes, int timeout)
        {
            var startTime = DateTime.Now;
            byte[] frame = new byte[4096];

            while ((DateTime.Now - startTime).TotalMilliseconds < timeout)
            {
                if (COM.IsOpen)
                {
                    while (COM.BytesToRead > 0)
                    {
                        int data = COM.ReadByte();
                        if (data == -1)
                            break;

                        if (_bufferIndex < _buffer.Length)
                        {
                            _buffer[_bufferIndex] = (byte)data;
                            _bufferIndex++;
                        }
                    }
                    frame = ExtractFrame(startBytes);
                    if (frame != null)
                    {
                        return frame;
                    }
                }
                if (Client.Connected)
                {
                    while (netStream.DataAvailable)
                    {
                        int data = netStream.ReadByte();
                        if (data == -1)
                            break;

                        if (_bufferIndex < _buffer.Length)
                        {
                            _buffer[_bufferIndex] = (byte)data;
                            _bufferIndex++;
                        }
                    }

                    // Try to extract frame
                    frame = ExtractFrame(startBytes);
                    if (frame != null)
                    {
                        return frame;
                    }
                }
            }
            frame = new byte[5];
            frame[0] = 0x01;
            frame[1] = 0x02;
            frame[2] = 0xCC;
            frame[3] = 0x20;
            frame[4] = 0xEF;
            return frame;
        }

        private async Task<byte[]> WaitForResponseAsync(byte[] startBytes, int timeout)
        {
            var startTime = DateTime.Now;
            byte[] frame = new byte[125];

            while ((DateTime.Now - startTime).TotalMilliseconds < timeout)
            {
                if (COM.IsOpen)
                {
                    while (COM.BytesToRead > 0)
                    {
                        int data = COM.ReadByte();
                        if (data == -1)
                            break;

                        if (_bufferIndex < _buffer.Length)
                        {
                            _buffer[_bufferIndex] = (byte)data;
                            _bufferIndex++;
                        }
                    }
                    frame = ExtractFrame(startBytes);
                    if (frame != null)
                    {
                        return frame;
                    }
                }
                if (Client.Connected)
                {

                    while (netStream.DataAvailable)
                    {
                        byte[] buffer2 = new byte[1];
                        int data = await netStream.ReadAsync(buffer2, 0, 1);
                        if (data == 0)
                            break;

                        if (_bufferIndex < _buffer.Length)
                        {
                            _buffer[_bufferIndex] = buffer2[0];
                            _bufferIndex++;
                        }

                    }

                    frame = ExtractFrame(startBytes);
                    if (frame != null)
                    {
                        return frame;
                    }
                }
                await Task.Delay(1);
            }
            frame = new byte[125];
            frame[0] = 0x01;
            frame[1] = 0x02;
            frame[2] = 0xCC;
            frame[3] = 0x20;
            frame[4] = 0xEF;
            return frame;
        }

        private byte sendCommand(byte cmd)
        {
            byte ret = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[4] { 0x01, 0x02, cmd, 0x00 };
                data[3] = getCRC(data, 3);
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _timeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                            return 0;
                        else
                            return errorCode;
                    }
                    return ret;
                }
                else
                {
                    errorCode = 0xFE;
                    return 0xFE;
                }
            }
            else
            {
                errorCode = 0xFF;
                return 0xFF;
            }
        }

        private async Task<byte> sendCommandMotionAsync(byte cmd)
        {
            byte rn = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[4] { 0x01, 0x02, cmd, 0x00 };
                data[3] = getCRC(data, 3);
                int ret = 0;
                if (sendData(data))
                {
                    do
                    {
                        byte[] res = await WaitForResponseAsync(startBytes, _motionTimeOut);
                        rn = ValidateResponse(res);
                        if (rn == 0)
                        {
                            ret = DecodeResponse(res);
                            if (ret == 2)
                                return 0;
                            if (ret == -1)
                                return errorCode;
                        }
                        else return rn;
                    }
                    while (true);
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }

        private async Task<byte> sendCommandMotionAsync(byte cmd, byte spd, byte acc, byte dec)
        {
            byte rn = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[7] { 0x01, 0x02, cmd, spd, acc, dec, 0x00 };
                data[6] = getCRC(data, 6);
                int ret = 0;
                if (sendData(data))
                {
                    byte[] res = await WaitForResponseAsync(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else
                        return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }

        private byte sendCommandMotion(byte cmd)
        {
            byte rn = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[4] { 0x01, 0x02, cmd, 0x00 };
                data[3] = getCRC(data, 3);
                int ret = 0;
                if (sendData(data))
                {
                    do
                    {
                        byte[] res = WaitForResponse(startBytes, _motionTimeOut);
                        rn = ValidateResponse(res);
                        if (rn == 0)
                        {
                            ret = DecodeResponse(res);
                            if (ret == 2)
                                return 0;
                            if (ret == -1)
                                return errorCode;
                        }
                        else return rn;
                    }
                    while (true);
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }

        private byte sendCommandMotion(byte cmd, byte spd, byte acc, byte dec)
        {
            byte rn = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[7] { 0x01, 0x02, cmd, spd, acc, dec, 0x00 };
                data[6] = getCRC(data, 6);
                int ret = 0;
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }

        private static string[] FindTeensyPorts(string deviceIdSubstring)
        {
            if (RuntimeInformation.IsOSPlatform(OSPlatform.Windows))
            {
                return FindTeensyPortsWindows(deviceIdSubstring);
            }
            else if (RuntimeInformation.IsOSPlatform(OSPlatform.Linux) || RuntimeInformation.IsOSPlatform(OSPlatform.OSX))
            {
                return FindTeensyPortsUnix("0478");
            }
            else
            {
                throw new PlatformNotSupportedException("Unsupported platform");
            }
        }

        private static string[] FindTeensyPortsWindows(string deviceIdSubstring)
        {
            List<string> ports = new List<string>();

            using (var searcher = new System.Management.ManagementObjectSearcher(
                "SELECT * FROM Win32_PnPEntity WHERE Caption LIKE '%(COM%'"))
            {
                foreach (var obj in searcher.Get())
                {
                    string deviceId = obj["DeviceID"]?.ToString();
                    string caption = obj["Caption"]?.ToString();

                    if (!string.IsNullOrEmpty(deviceId) && deviceId.Contains(deviceIdSubstring))
                    {
                        string comPort = caption.Substring(caption.LastIndexOf("(COM")).Replace("(", "").Replace(")", "");
                        ports.Add(comPort);
                    }
                }
            }

            return ports.ToArray();
        }

        private static string[] FindTeensyPortsUnix(string deviceIdSubstring)
        {
            List<string> ports = new List<string>();

            string[] devicePaths = Directory.GetDirectories("/sys/bus/usb/devices/");
            foreach (string devicePath in devicePaths)
            {
                try
                {
                    string idProductPath = Path.Combine(devicePath, "idProduct");
                    if (File.Exists(idProductPath))
                    {
                        string idProduct = File.ReadAllText(idProductPath).Trim();
                        if (idProduct == deviceIdSubstring)
                        {
                            // Find the tty device under this USB device path
                            string[] ttyDevices = Directory.GetFiles(devicePath, "tty*");
                            foreach (string ttyDevice in ttyDevices)
                            {
                                string port = Path.Combine("/dev", Path.GetFileName(ttyDevice));
                                ports.Add(port);
                            }
                        }
                    }
                }
                catch (Exception ex)
                {
                    return null;
                }
            }

            return ports.ToArray();
        }
        private void CloseConnection()
        {
            if (netStream != null)
            {
                netStream.Close();
                netStream.Dispose();
                netStream = null;
            }
            if (Client != null)
            {
                Client.Close();
                Client.Dispose();
                Client = null;
            }
            if (COM.IsOpen)
                COM.Close();
        }

        private static bool IsValidIPAddress(string input)
        {
            if (string.IsNullOrEmpty(input))
                return false;

            string[] parts = input.Split('.');
            if (parts.Length != 4)
                return false;

            foreach (string part in parts)
            {
                if (!int.TryParse(part, out int num))
                    return false;

                if (num < 0 || num > 255)
                    return false;
            }

            return true;
        }

        /*--------------------------- Common methods ------------------------------- */
        public string getLibVersion()
        {
            return _version;
        }
        public List<string> findUSB()
        {
            string teensyDeviceIdSubstring = "USB\\VID_16C0&PID_048";
            string[] comPorts = FindTeensyPorts(teensyDeviceIdSubstring);
            List<string> ports = new List<string>();
            ports.AddRange(comPorts);
            if (comPorts.Length > 0)
                return ports;
            else
                return null;
        }

        public async Task<List<string>> findEthernet()
        {
            List<string> ips = new List<string>();
            int i = 0;
            try
            {
                IReadOnlyList<IZeroconfHost> results = await ZeroconfResolver.ResolveAsync("astorino.local.");

                foreach (var host in results)
                {
                    foreach (var ip in host.IPAddresses)
                    {
                        ips.Add(ip);
                    }
                }
            }
            catch
            {
                return null;
            }

            return ips;

        }
        private byte ConnectUSB(string portName)
        {
            if (COM.IsOpen)
                COM.Close();
            COM = new SerialPort(portName, 256000);
            COM.DtrEnable = true;
            try
            {
                COM.Open();
            }
            catch
            {
                return 0xFF;
            }
            return sendCommand(0x24);
        }


        public void setTimeOut(int timeOut)
        {
            _timeOut = timeOut;
        }

        public void setMotionTimeOut(int timeOut)
        {
            _motionTimeOut = timeOut;
        }

        public byte ConnectViaUART(string portName)
        {
            if (COM.IsOpen)
                COM.Close();
            COM = new SerialPort(portName, 115200);
            try
            {
                COM.Open();
            }
            catch
            {
                return 0xFF;
            }
            return sendCommand(0x19);
        }

        public byte Connect(string path)
        {
            if (path.Contains("COM") || path.Contains("com"))
            {
                return ConnectUSB(path);
            }
            else
            {
                if (IsValidIPAddress(path))
                {
                    CloseConnection();
                    Client = new TcpClient();
                    Client.Connect(IPAddress.Parse(path), 23);
                    netStream = Client.GetStream();
                    return sendCommand(0x24);
                }
            }
            return 0xFF;
        }
        public byte Disconnect()
        {
            if (COM.IsOpen || Client != null)
            {
                if (sendCommand(0x25) == 0)
                {
                    if (COM.IsOpen)
                        COM.Close();
                    return 0;
                }
                else
                    return errorCode;
            }
            else
                return 0xFF;
        }

        public byte cancelMotion()
        {
            return sendCommand(0x45);
        }

        public byte emergencyStop()
        {
            return sendCommand(0x5A);
        }

        /* ----------------------------- matrix methods ----------------------------- */
        public retVal ForwardKinematics(double jt1, double jt2, double jt3, double jt4, double jt5, double jt6, double jt7)
        {
            byte ret = 0;
            Int32 pose = 0;
            double[] jt = new double[7] { jt1, jt2, jt3, jt4, jt5, jt6, jt7 };
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[32];
                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x54;
                for (int i = 0; i < 7; i++)
                {
                    pose = (Int32)(jt[i] * 1000.0);

                    data[3 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[4 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[5 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[6 + i * 4] = (byte)(pose & 0xFF);
                }

                data[31] = getCRC(data, 31);
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _timeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                            return new retVal(forw, 0);
                        else
                            return new retVal(errorCode);
                    }
                    return new retVal(errorCode);
                }
                else
                    return new retVal(0xFE);
            }
            else
                return new retVal(0xFF);
        }

        public retVal ForwardKinematics(double jt1, double jt2, double jt3, double jt4, double jt5, double jt6)
        {
            return ForwardKinematics(jt1, jt2, jt3, jt4, jt5, jt6, 0);
        }

        public retVal ForwardKinematics(double[] jt)
        {
            if (jt.Length == 6)
                return ForwardKinematics(jt[0], jt[1], jt[2], jt[3], jt[4], jt[5], 0);
            if (jt.Length == 7)
                return ForwardKinematics(jt[0], jt[1], jt[2], jt[3], jt[4], jt[5], jt[6]);

            return new retVal(0x22);
        }
        public retVal InverseKinematics(double x, double y, double z, double o, double a, double t, double jt7)
        {
            byte ret = 0;
            Int32 pose = 0;
            double[] xyz = new double[7] { x, y, z, o, a, t, jt7 };
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[32];
                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x55;
                for (int i = 0; i < 7; i++)
                {
                    pose = (Int32)(xyz[i] * 1000.0);

                    data[3 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[4 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[5 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[6 + i * 4] = (byte)(pose & 0xFF);
                }

                data[31] = getCRC(data, 31);
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _timeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                            return new retVal(inv, 0);
                        else
                            return new retVal(errorCode);
                    }
                    return new retVal(errorCode);
                }
                else 
                    return new retVal(0xFE);
            }
            else
                return new retVal(0xFF);
        }

        public retVal InverseKinematics(double x, double y, double z, double o, double a, double t)
        {
            return InverseKinematics(x, y, z, o, a, t, 0);
        }

        public retVal InverseKinematics(double[] pose)
        {
            if (pose.Length == 6)
                return InverseKinematics(pose[0], pose[1], pose[2], pose[3], pose[4], pose[5], 0.0);
            if (pose.Length == 7)
                return InverseKinematics(pose[0], pose[1], pose[2], pose[3], pose[4], pose[5], pose[6]);

            return new retVal(0x22);
        }

        public retVal toOAT(double rx, double ry, double rz)
        {
            byte ret = 0;
            Int32 pose = 0;
            double[] jt = new double[3] { rx, ry, rz };
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[16];
                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x58;
                for (int i = 0; i < 3; i++)
                {
                    pose = (Int32)(jt[i] * 1000.0);

                    data[3 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[4 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[5 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[6 + i * 4] = (byte)(pose & 0xFF);
                }

                data[15] = getCRC(data, 15);
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _timeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                            return new retVal(oat, 0);
                        else
                            return new retVal(errorCode);
                    }
                    return new retVal(errorCode);
                }
                else
                   return new retVal(0xFE);
            }
            else
                return new retVal(0xFF);
        }

        public retVal toRPY(double o, double a, double t)
        {
            byte ret = 0;
            Int32 pose = 0;
            double[] jt = new double[3] { o, a, t };
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[16];
                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x59;
                for (int i = 0; i < 3; i++)
                {
                    pose = (Int32)(jt[i] * 1000.0);

                    data[3 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[4 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[5 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[6 + i * 4] = (byte)(pose & 0xFF);
                }

                data[15] = getCRC(data, 15);
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _timeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                            return new retVal(euler, 0);
                        else
                            return new retVal(errorCode);
                    }
                    return new retVal(errorCode);
                }
                else
                    return new retVal(0xFE);
            }
            else
                return new retVal(0xFF);
        }

        /* ----------------------------- set astorino status data ------------------*/
        public byte setMotorOn()
        {
            return sendCommand(0x20);
        }
        public byte setMotorOff()
        {
            byte ret = 0;
            if (sendCommand(0x21) == 0)
            {
                if (COM.IsOpen || Client.Connected)
                {
                    byte[] data = new byte[4] { 0x01, 0x02, 0x21, 0x00 };
                    data[3] = getCRC(data, 3);
                    if (sendData(data))
                    {
                        if (IsBitSet(status[1], 1))
                        {
                            //zeroing done - wait for motion completed
                            byte[] res = WaitForResponse(startBytes, _motionTimeOut);
                            ret = ValidateResponse(res);
                            if (ret == 0)
                            {
                                if (DecodeResponse(res) >= 0)
                                    return 0;
                                else
                                    return errorCode;
                            }
                            else
                                return ret;
                        }
                        else
                        {
                            //zeroing not done - wait for command success
                            byte[] res = WaitForResponse(startBytes, _timeOut);
                            ret = ValidateResponse(res);
                            if (ret == 0)
                            {
                                if (DecodeResponse(res) >= 0)
                                    return 0;
                                else
                                    return errorCode;
                            }
                            return ret;
                        }
                    }
                    else
                        return 0xFE;
                }
                else
                    return 0xFF;
            }
            else
                return errorCode;
        }
        public byte reset()
        {
            errorCode = 0x00;
            return sendCommand(0x22);
        }

        public byte Hold()
        {
            return sendCommand(0x30);
        }

        public byte Run()
        {
            return sendCommand(0x31);
        }
        public byte turnDryRunOn()
        {
            return sendCommand(0x32);
        }
        public byte turnDryRunOff()
        {
            return sendCommand(0x33);
        }
        public byte setRepeatCont()
        {
            return sendCommand(0x34);
        }
        public byte setRepeatOnce()
        {
            return sendCommand(0x35);
        }
        public byte setStepOnce()
        {
            return sendCommand(0x36);
        }
        public byte setStepCont()
        {
            return sendCommand(0x37);
        }

        public byte cycleStop()
        {
            return sendCommand(0x3C);
        }
        public byte cycleStart()
        {
            return sendCommand(0x2B);
        }

        public byte nextStep()
        {
            return sendCommand(0x6C );
        }
        public byte skipWait()
        {
            return sendCommand(0x6D);
        }

        /* ------------------------ IOs -----------------------------*/

        public retVal readInput(int ind)
        {
            if (ind <= 0 || ind > 58)
            {
                return new retVal(0x08);
            }

            ind = ind - 1;
            int count = ind / 8;
            if (sendCommand(0x26) == 0)
            {
                if (IsBitSet(IO[count], ind % 8))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
            {
                return new retVal(errorCode);
            }
        }
        public retVal readOutput(int ind)
        {
            if (ind <= 0 || ind > 58)
                return new retVal(0x08);

            ind = ind - 1;
            int count = ind / 8;
            if (sendCommand(0x26) == 0)
            {
                if (IsBitSet(IO[count + 8], ind % 8))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
            {
                return new retVal(errorCode);
            }
        }
        public retVal readInternal(int ind)
        {
            if (ind <= 0 || ind > 16)
                return new retVal(0x08);
            if (sendCommand(0x26) == 0)
            {
                if (ind < 9)
                {
                    if (IsBitSet(IO[17], ind - 1))
                        return new retVal(1, 0);
                    else
                        return new retVal(-1, 0);
                }
                else
                {
                    if (IsBitSet(IO[16], ind - 9))
                        return new retVal(1, 0);
                    else
                        return new retVal(-1, 0);
                }
            }
            else
            {
                return new retVal(errorCode);
            }
        }

        public byte setOutput(int output, int state)
        {
            byte ret = 0;
            if (COM.IsOpen || Client.Connected)
            {
                if (state == -1)
                    state = 0;
                byte[] data = new byte[6] { 0x01, 0x02, 0x3A, (byte)output, (byte)state, 0x00 };
                data[5] = getCRC(data, 5);
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _timeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                            return ret;
                        else
                            return errorCode;
                    }
                    return ret;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }

        public byte setInternal(int sig, int state)
        {
            byte ret = 0;
            if (COM.IsOpen || Client.Connected)
            {
                if (state == -1)
                    state = 0;
                byte[] data = new byte[6] { 0x01, 0x02, 0x4C, (byte)sig, (byte)state, 0x00 };
                data[5] = getCRC(data, 5);
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _timeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                            return ret;
                        else
                            return errorCode;
                    }
                    return ret;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }

        /* ------------------------- read astorino status ------------------------- */

        public retVal readStatusBytes()
        {
            if (sendCommand(0x27) == 0)
                return new retVal(status, 0);
            else
                return new retVal(errorCode);
        }
        public retVal isMotorOn()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[0], 6))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal isInHome()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[0], 7))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal isRepeatModeOn()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[0], 5))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal isHoldOn()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[0], 4))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal isCycleOn()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[0], 3))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal isEstopOn()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[0], 2))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal isErrorOn()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[0], 1))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal isReadyOn()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[0], 0))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal isExternalHoldOn()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[1], 7))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal isSafetyFenceOn()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[1], 6))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal isRepeatContOn()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[1], 5))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal isStepOnceOn()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[1], 4))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal isStepWaitingOn()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[1], 3))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal isDryRunOn()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[1], 2))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal isZeroingDone()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[1], 1))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal isInMotion()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[1], 0))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal isIOActive()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[2], 7))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }

        public retVal isModbusConnected()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[3], 1))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }

        public retVal isCollisionDetectionActive()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[3], 0))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal isZeroingRunning()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[4], 6))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }

        public retVal isTeachMotionActive()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[4], 5))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }

        public retVal isMotionCommand()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[4], 4))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }

        public retVal readSelectedTeachMotionMode()
        {
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[4], 0))
                    return new retVal(1, 0); // tool
                if (IsBitSet(status[4], 1))
                    return new retVal(2, 0); // joint
                if (IsBitSet(status[4], 2))
                    return new retVal(3, 0); // conv
                if (IsBitSet(status[4], 3))
                    return new retVal(4, 0); // base
                if (IsBitSet(status[4], 7))
                    return new retVal(5, 0); // work

                return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }

        public retVal readSelectedToolNumber()
        {
            if (sendCommand(0x27) == 0)
            {
                int num = ConvertBitsToInt(IsBitSet(status[3], 5), IsBitSet(status[3], 6), IsBitSet(status[3], 7)) + 1;
                return new retVal(num, 0);
            }
            else
                return new retVal(errorCode);
        }

        public retVal readSelectedTeachSpeed()
        {
            if (sendCommand(0x27) == 0)
            {
                int num = ConvertBitsToInt(IsBitSet(status[3], 2), IsBitSet(status[3], 3), IsBitSet(status[3], 4)) + 1;
                return new retVal(num, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal isHallSensorOn(int jt)
        {
            if (jt < 1 || jt > 7)
                return new retVal(0x08);
            if (sendCommand(0x27) == 0)
            {
                if (IsBitSet(status[2], (jt - 1)))
                    return new retVal(1, 0);
                else
                    return new retVal(-1, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal JT()
        {
            if (sendCommand(0x28) == 0)
            {
                return new retVal(joints, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal JT_Velocity()
        {
            if (sendCommand(0x67) == 0)
            {
                return new retVal(jt_vel, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal Pose()
        {
            if (sendCommand(0x29) == 0)
            {
                return new retVal(pose, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal TCP_Velocity()
        {
            if (sendCommand(0x66) == 0)
            {
                return new retVal(abs_vel, 0);
            }
            else
                return new retVal(errorCode);
        }

        public retVal AccelerometerData()
        {
            if (sendCommand(0x6E) == 0)
            {
                return new retVal(accData, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal cpuTemp()
        {
            if (sendCommand(0x5B) == 0)
            {
                return new retVal(cpu_temp, 0);
            }
            else
                return new retVal(errorCode);
        }

        public retVal readMonitorSpeed()
        {
            if (sendCommand(0x5E) == 0)
            {
                return new retVal(rS, 0);
            }
            else
                return new retVal(errorCode);
        }

        public retVal readErrorCode()
        {
            if (sendCommand(0x62) == 0)
            {
                return new retVal(errCode, 0);
            }
            else
                return new retVal(errorCode);
        }
        /* ----------------------------- read astoriono data -------------------------- */
        public retVal serialNumber()
        {
            if (sendCommand(0x2A) == 0)
            {
                return new retVal(sn, 0);
            }
            else
                return new retVal(errorCode);
        }



        public retVal firmwareVersion()
        {
            if (sendCommand(0x38) == 0)
            {
                return new retVal(frmV, 0);
            }
            else
                return new retVal(errorCode);
        }

        public TransformationPoints readTransPointsData()
        {
            byte ret = 0;
            int rn = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[4] { 0x01, 0x02, 0x3D, 0x00 };
                data[3] = getCRC(data, 3);
                if (sendData(data))
                {
                    do
                    {
                        data = new byte[4] { 0x01, 0x02, 0x06, 0x00 };
                        data[3] = getCRC(data, 3);
                        byte[] res = WaitForResponse(startBytes, _timeOut);
                        ret = ValidateResponse(res);
                        if (ret == 0)
                        {
                            rn = DecodeResponse(res);
                            if (rn == 1)
                                break;
                            if (rn == -1)
                                new TransformationPoints(errorCode);
                            sendData(data);
                        }
                        else
                            return new TransformationPoints(errorCode);
                    } while (true);
                }
                else
                    return new TransformationPoints(errorCode);

                return new TransformationPoints(transPoints, 0);
            }
            else
                return new TransformationPoints(errorCode);
        }
        public JointPoints readJointPointsData()
        {
            byte ret = 0;
            int rn = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[4] { 0x01, 0x02, 0x3E, 0x00 };
                data[3] = getCRC(data, 3);
                if (sendData(data))
                {
                    data = new byte[4] { 0x01, 0x02, 0x06, 0x00 };
                    data[3] = getCRC(data, 3);
                    do
                    {
                        byte[] res = WaitForResponse(startBytes, _timeOut);
                        ret = ValidateResponse(res);
                        if (ret == 0)
                        {
                            rn = DecodeResponse(res);
                            if (rn == 1)
                                break;
                            if (rn == -1)
                                return new JointPoints(errorCode);
                            sendData(data);
                        }
                        else
                            return new JointPoints(errorCode);
                    } while (true);
                }
                else
                    return new JointPoints(errorCode);

                return new JointPoints(jointPoints, 0);
            }
            else
                return new JointPoints(errorCode);
        }

        /* ---------------------------- set astorino data ------------------------*/
        public byte setTool(int index)
        {
            if (index == 1)
                return sendCommand(0x2C);
            else if (index == 2)
                return sendCommand(0x2D);
            else if (index == 2)
                return sendCommand(0x2E);
            else
                return 0x08;
        }

        public byte setWork(int index)
        {
            if (index == 1)
                return sendCommand(0x63);
            else if (index == 2)
                return sendCommand(0x64);
            else
                return 0x08;
        }
        public byte setHomeHere()
        {
            return sendCommand(0x5D);
        }
        public byte setUartTimeout(int timeout)
        {
            byte ret = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[5] { 0x01, 0x02, 0x6F, (byte)timeout, 0x00 };
                data[4] = getCRC(data, 4);
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _timeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                        {
                            return ret;
                        }
                        else
                            return errorCode;
                    }
                    return ret;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public byte setMonitorSpeed(int speed)
        {
            byte ret = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[5] { 0x01, 0x02, 0x39, (byte)speed, 0x00 };
                data[4] = getCRC(data, 4);
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _timeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                        {
                            rS = speed;
                            return ret;
                        }
                        else
                            return errorCode;
                    }
                    return ret;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public byte setHome(double jt1, double jt2, double jt3, double jt4, double jt5, double jt6, double jt7)
        {
            byte ret = 0;
            Int32 pose = 0;
            double[] jt = new double[7] { jt1, jt2, jt3, jt4, jt5, jt6, jt7 };
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[32];
                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x23;
                for (int i = 0; i < 7; i++)
                {
                    pose = (Int32)(jt[i] * 1000.0);

                    data[3 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[4 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[5 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[6 + i * 4] = (byte)(pose & 0xFF);
                }

                data[31] = getCRC(data, 31);
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _timeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                            return ret;
                        else
                            return errorCode;
                    }
                    return ret;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }

        public byte setHome(double jt1, double jt2, double jt3, double jt4, double jt5, double jt6)
        {
            return setHome(jt1, jt2, jt3, jt4, jt5, jt6, 0.0);
        }

        public byte setHome(double[] jt)
        {
            if (jt.Length == 6)
                return setHome(jt[0], jt[1], jt[2], jt[3], jt[4], jt[5], 0.0);
            if (jt.Length == 7)
                return setHome(jt[0], jt[1], jt[2], jt[3], jt[4], jt[5], jt[6]);

            return 0x22;
        }
        public byte saveCurrentJointsAsPoint(int point)
        {
            double[] joint = new double[7] { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
            int i = 0;
            byte ret = 0;
            retVal temp;
            if (COM.IsOpen || Client.Connected)
            {
                if (point < 0) return 0x08;
                if (point > 99) return 0x02;
                temp = JT();
                if (temp.ReturnCode == 0)
                {
                    for (i = 0; i < 7; i++)
                        jointPoints[point, i] = temp.Values[i];//joint[i];
                    byte[] data = new byte[5] { 0x01, 0x02, 0x3F, (byte)point, 0x00 };
                    data[4] = getCRC(data, 4);
                    if (sendData(data))
                    {
                        byte[] res = WaitForResponse(startBytes, _timeOut);
                        ret = ValidateResponse(res);
                        if (ret == 0)
                        {
                            if (DecodeResponse(res) >= 0)
                                return ret;
                            else
                                return errorCode;
                        }
                        return ret;
                    }
                    else
                        return 0xFE;
                }
                else
                    return temp.ReturnCode;
            }
            else
                return 0xFF;
        }
        public byte saveCurrentPoseAsPoint(int point)
        {
            double[] pose = new double[7] { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
            int i = 0;
            byte ret = 0;
            retVal temp;
            if (COM.IsOpen || Client.Connected)
            {
                temp = Pose();
                if (temp.ReturnCode == 0)
                {
                    for (i = 0; i < 7; i++)
                        transPoints[point, i] = temp.Values[i];
                    byte[] data = new byte[5] { 0x01, 0x02, 0x41, (byte)point, 0x00 };
                    data[4] = getCRC(data, 4);
                    if (sendData(data))
                    {
                        byte[] res = WaitForResponse(startBytes, _timeOut);
                        ret = ValidateResponse(res);
                        if (ret == 0)
                        {
                            if (DecodeResponse(res) >= 0)
                                return ret;
                            else
                                return errorCode;
                        }
                        return ret;
                    }
                    else
                        return 0xFE;
                }
                else
                    return temp.ReturnCode;
            }
            else
                return 0xFF;
        }
        public byte setMainProgram(string name)
        {
            byte ret = 0;
            if (COM.IsOpen || Client.Connected)
            {
                if (SendFrame(name, 0x43))
                {
                    byte[] res = WaitForResponse(startBytes, _timeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                            return ret;
                        else
                            return errorCode;
                    }
                    return ret;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }

        public byte selectProgram(string name)
        {
            byte ret = 0;
            if (COM.IsOpen || Client.Connected)
            {
                if (SendFrame(name, 0x44))
                {
                    byte[] res = WaitForResponse(startBytes, _timeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                            return ret;
                        else
                            return errorCode;
                    }
                    return ret;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }

        public retVal readProgramsName()
        {
            byte ret = 0;
            int rn = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[4] { 0x01, 0x02, 0x46, 0x00 };
                data[3] = getCRC(data, 3);
                if (sendData(data))
                {
                    data = new byte[4] { 0x01, 0x02, 0x06, 0x00 };
                    data[3] = getCRC(data, 3);
                    do
                    {
                        byte[] res = WaitForResponse(startBytes, _timeOut);
                        ret = ValidateResponse(res);
                        if (ret == 0)
                        {
                            rn = DecodeResponse(res);
                            if (rn == 1)
                                break;
                            if (rn == -1)
                                return new retVal(errorCode);
                            sendData(data);
                        }
                        else
                            return new retVal(errorCode);
                    } while (true);
                }
                else
                    return new retVal(0xFE);

                return new retVal(programsName, 0);
            }
            else
                return new retVal(errorCode);
        }

        public retVal mainProgram()
        {
            if (sendCommand(0x47) == 0)
            {
                return new retVal(mP, 0);
            }
            else
                return new retVal(errorCode);
        }
        public retVal selectedProgram()
        {
            if (sendCommand(0x57) == 0)
            {
                return new retVal(sP, 0);
            }
            else
                return new retVal(errorCode);
        }

        public byte sendToolData(int index, double[] data)
        {
            byte ret = 0;
            if (COM.IsOpen || Client.Connected)
            {
                if (SendData((byte)index, 0x48, data))
                {
                    byte[] res = WaitForResponse(startBytes, _timeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                            return ret;
                        else
                            return errorCode;
                    }
                    return ret;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public byte removePoint(int number, int type)
        {
            byte ret = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[6] { 0x01, 0x02, 0x49, (byte)number, (byte)type, 0x00 };
                data[5] = getCRC(data, 5);
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _timeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                            return ret;
                        else
                            return errorCode;
                    }
                    return ret;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public retVal readToolData(int index)
        {
            byte ret = 0;
            if (index < 1 || index > 3)
                return new retVal(0x08);

            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[5] { 0x01, 0x02, 0x56, (byte)index, 0x00 };
                data[4] = getCRC(data, 4);
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _timeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                        {
                            if (index == 1)
                                return new retVal(Tool1, 0);
                            else if (index == 2)
                                return new retVal(Tool2, 0);
                            else if (index == 3)
                                return new retVal(Tool3, 0);
                            else
                                return new retVal(0x08);
                        }
                        else
                            return new retVal(errorCode);
                    }
                }
                else
                    return new retVal(0xFE);

                return new retVal(errorCode);
            }
            else
                return new retVal(errorCode);
        }

        public retVal readTransformationPoint(int index)
        {
            byte ret = 0;
            if (index < 0 || index > 99)
                return new retVal(0x08);

            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[5] { 0x01, 0x02, 0x5F, (byte)index, 0x00 };
                data[4] = getCRC(data, 4);
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _timeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                        {
                            double[] arr = new double[7];
                            for (int i = 0; i < 7; i++)
                                arr[i] = transPoints[index, i];
                            return new retVal(arr, 0);
                        }
                        else
                            return new retVal(errorCode);
                    }
                }
                else
                    return new retVal(0xFE);

                return new retVal(errorCode);
            }
            else
                return new retVal(errorCode);
        }
        public retVal readJointPoint(int index)
        {
            byte ret = 0;
            if (index < 0 || index > 99)
                return new retVal(0x08);

            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[5] { 0x01, 0x02, 0x60, (byte)index, 0x00 };
                data[4] = getCRC(data, 4);
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _timeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                        {
                            double[] arr = new double[7];
                            for (int i = 0; i < 7; i++)
                                arr[i] = jointPoints[index, i];
                            return new retVal(arr, 0);
                        }
                        else
                            return new retVal(errorCode);
                    }
                }
                else
                    return new retVal(0xFE);

                return new retVal(errorCode);
            }
            else
                return new retVal(errorCode);
        }

        /* ------------------------ Synch motion commands -------------------------------- */
        public byte Zero()
        {
            return sendCommandMotion(0x3B);
        }
        public byte JMOVE(byte pointType, byte pointIndex, byte spd, byte acc, byte dec)
        {
            byte rn = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[9] { 0x01, 0x02, 0x4F, pointType, pointIndex, spd, acc, dec, 0x00 };
                data[8] = getCRC(data, 8);
                int ret = 0;
                if (sendData(data))
                {

                    byte[] res = WaitForResponse(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else return 0xFE;
            }
            else
                return 0xFF;
        }
        public byte JMOVE(byte pointType, byte spd, byte acc, byte dec, double[] target)
        {
            byte rn = 0;
            Int32 pose = 0;
            int j = target.Length;
            if (j < 6)
                return 0x22;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[36];
                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x50;
                data[3] = pointType;
                data[4] = spd;
                data[5] = acc;
                data[6] = dec;

                for (int i = 0; i < j; i++)
                {
                    pose = (Int32)(target[i] * 1000.0);

                    data[7 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[8 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[9 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[10 + i * 4] = (byte)(pose & 0xFF);
                }
                if (j == 6)
                {
                    data[31] = 0;
                    data[32] = 0;
                    data[33] = 0;
                    data[34] = 0;
                }
                data[35] = getCRC(data, 35);
                int ret = 0;
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public byte LMOVE(byte pointType, byte pointIndex, byte spd, byte acc, byte dec)
        {
            byte rn = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[9] { 0x01, 0x02, 0x4D, pointType, pointIndex, spd, acc, dec, 0x00 };
                data[8] = getCRC(data, 8);
                int ret = 0;
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public byte LMOVE(byte pointType, byte spd, byte acc, byte dec, double[] target)
        {
            byte rn = 0;
            Int32 pose = 0;
            int j = target.Length;
            if (j < 6)
                return 0x22;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[36];
                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x4E;
                data[3] = pointType;
                data[4] = spd;
                data[5] = acc;
                data[6] = dec;

                for (int i = 0; i < j; i++)
                {
                    pose = (Int32)(target[i] * 1000.0);

                    data[7 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[8 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[9 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[10 + i * 4] = (byte)(pose & 0xFF);
                }
                if (j == 6)
                {
                    data[31] = 0;
                    data[32] = 0;
                    data[33] = 0;
                    data[34] = 0;
                }
                data[35] = getCRC(data, 35);
                int ret = 0;
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public byte HOME(byte spd, byte acc, byte dec)
        {
            return sendCommandMotion(0x2F, spd, acc, dec);
        }
        public byte CMOVE(byte pointType, byte spd, byte acc, byte dec, double[] middle, double[] target)
        {
            byte rn = 0;
            Int32 pose = 0;
            int j = target.Length;
            if (j < 6)
                return 0x22;

            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[64];
                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x61;
                data[3] = pointType;
                data[4] = spd;
                data[5] = acc;
                data[6] = dec;

                for (int i = 0; i < j; i++)
                {
                    pose = (Int32)(middle[i] * 1000.0);

                    data[7 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[8 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[9 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[10 + i * 4] = (byte)(pose & 0xFF);
                }
                if (j == 6)
                {
                    data[31] = 0;
                    data[32] = 0;
                    data[33] = 0;
                    data[34] = 0;
                }

                for (int i = 0; i < j; i++)
                {
                    pose = (Int32)(target[i] * 1000.0);

                    data[35 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[36 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[37 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[38 + i * 4] = (byte)(pose & 0xFF);
                }
                if (j == 6)
                {
                    data[59] = 0;
                    data[60] = 0;
                    data[61] = 0;
                    data[62] = 0;
                }

                data[63] = getCRC(data, 63);
                int ret = 0;
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public byte CMOVE(byte pointType1, byte pointIndex1, byte pointIndex2, byte spd, byte acc, byte dec)
        {
            byte rn = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[10] { 0x01, 0x02, 0x65, pointType1, pointIndex1, pointIndex2, spd, acc, dec, 0x00 };
                data[9] = getCRC(data, 9);
                int ret = 0;
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public byte JAPPRO(byte pointType, byte pointIndex, byte spd, byte acc, byte dec, double offset)
        {
            byte rn = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[13];
                int ret = 0;
                Int32 pose = 0;

                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x68;
                data[3] = pointType;
                data[4] = pointIndex;
                data[5] = spd;
                data[6] = acc;
                data[7] = dec;

                pose = (Int32)(offset * 1000.0);

                data[8] = (byte)((pose >> 24) & 0xFF);
                data[9] = (byte)((pose >> 16) & 0xFF);
                data[10] = (byte)((pose >> 8) & 0xFF);
                data[11] = (byte)(pose & 0xFF);

                data[12] = getCRC(data, 12);
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public byte LAPPRO(byte pointType, byte pointIndex, byte spd, byte acc, byte dec, double offset)
        {
            byte rn = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[13];
                int ret = 0;
                Int32 pose = 0;

                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x69;
                data[3] = pointType;
                data[4] = pointIndex;
                data[5] = spd;
                data[6] = acc;
                data[7] = dec;

                pose = (Int32)(offset * 1000.0);

                data[8] = (byte)((pose >> 24) & 0xFF);
                data[9] = (byte)((pose >> 16) & 0xFF);
                data[10] = (byte)((pose >> 8) & 0xFF);
                data[11] = (byte)(pose & 0xFF);

                data[12] = getCRC(data, 12);
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public byte LAPPRO(byte pointType, byte spd, byte acc, byte dec, double[] target, double offset)
        {
            byte rn = 0;
            Int32 pose = 0;
            int j = target.Length;
            if (j < 6)
                return 0x22;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[40];
                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x6A;
                data[3] = pointType;
                data[4] = spd;
                data[5] = acc;
                data[6] = dec;

                for (int i = 0; i < j; i++)
                {
                    pose = (Int32)(target[i] * 1000.0);

                    data[7 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[8 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[9 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[10 + i * 4] = (byte)(pose & 0xFF);
                }
                if (j == 6)
                {
                    data[31] = 0;
                    data[32] = 0;
                    data[33] = 0;
                    data[34] = 0;
                }

                pose = (Int32)(offset * 1000.0);

                data[35] = (byte)((pose >> 24) & 0xFF);
                data[36] = (byte)((pose >> 16) & 0xFF);
                data[37] = (byte)((pose >> 8) & 0xFF);
                data[38] = (byte)(pose & 0xFF);

                data[39] = getCRC(data, 39);

                int ret = 0;
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public byte JAPPRO(byte pointType, byte spd, byte acc, byte dec, double[] target, double offset)
        {
            byte rn = 0;
            Int32 pose = 0;
            int j = target.Length;
            if (j < 6)
                return 0x22;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[40];
                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x6B;
                data[3] = pointType;
                data[4] = spd;
                data[5] = acc;
                data[6] = dec;

                for (int i = 0; i < j; i++)
                {
                    pose = (Int32)(target[i] * 1000.0);

                    data[7 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[8 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[9 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[10 + i * 4] = (byte)(pose & 0xFF);
                }
                if (j == 6)
                {
                    data[31] = 0;
                    data[32] = 0;
                    data[33] = 0;
                    data[34] = 0;
                }

                pose = (Int32)(offset * 1000.0);

                data[35] = (byte)((pose >> 24) & 0xFF);
                data[36] = (byte)((pose >> 16) & 0xFF);
                data[37] = (byte)((pose >> 8) & 0xFF);
                data[38] = (byte)(pose & 0xFF);

                data[39] = getCRC(data, 39);

                int ret = 0;
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        /* ----------------------------- Async Motion commands ------------------------------*/

        public async Task<byte> ZeroAsync()
        {
            return await sendCommandMotionAsync(0x3B);
        }
        public async Task<byte> JMOVEAsync(byte pointType, byte pointIndex, byte spd, byte acc, byte dec)
        {
            byte rn = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[9] { 0x01, 0x02, 0x4F, pointType, pointIndex, spd, acc, dec, 0x00 };
                data[8] = getCRC(data, 8);
                int ret = 0;
                if (sendData(data))
                {
                    byte[] res = await WaitForResponseAsync(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public async Task<byte> LMOVEAsync(byte pointType, byte spd, byte acc, byte dec, double[] target)
        {
            byte rn = 0;
            Int32 pose = 0;
            int j = target.Length;
            if (j < 6)
                return 0x22;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[36];
                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x4E;
                data[3] = pointType;
                data[4] = spd;
                data[5] = acc;
                data[6] = dec;

                for (int i = 0; i < j; i++)
                {
                    pose = (Int32)(target[i] * 1000.0);

                    data[7 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[8 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[9 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[10 + i * 4] = (byte)(pose & 0xFF);
                }
                if (j == 6)
                {
                    data[31] = 0;
                    data[32] = 0;
                    data[33] = 0;
                    data[34] = 0;
                }
                data[35] = getCRC(data, 35);
                int ret = 0;
                if (sendData(data))
                {
                    byte[] res = await WaitForResponseAsync(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public async Task<byte> JMOVEAsync(byte pointType, byte spd, byte acc, byte dec, double[] target)
        {
            byte rn = 0;
            Int32 pose = 0;
            int j = target.Length;
            if (j < 6)
                return 0x22;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[36];
                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x50;
                data[3] = pointType;
                data[4] = spd;
                data[5] = acc;
                data[6] = dec;

                for (int i = 0; i < j; i++)
                {
                    pose = (Int32)(target[i] * 1000.0);

                    data[7 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[8 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[9 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[10 + i * 4] = (byte)(pose & 0xFF);
                }
                if (j == 6)
                {
                    data[31] = 0;
                    data[32] = 0;
                    data[33] = 0;
                    data[34] = 0;
                }
                data[35] = getCRC(data, 35);
                int ret = 0;
                if (sendData(data))
                {
                    byte[] res = await WaitForResponseAsync(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public async Task<byte> HOMEAsync(byte spd, byte acc, byte dec)
        {
            return await sendCommandMotionAsync(0x2F, spd, acc, dec);
        }
        public async Task<byte> CMOVEAsync(byte pointType, byte spd, byte acc, byte dec, double[] middle, double[] target)
        {
            byte rn = 0;
            Int32 pose = 0;
            int j = target.Length;
            if (j < 6)
                return 0x22;

            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[64];
                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x61;
                data[3] = pointType;
                data[4] = spd;
                data[5] = acc;
                data[6] = dec;

                for (int i = 0; i < j; i++)
                {
                    pose = (Int32)(middle[i] * 1000.0);

                    data[7 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[8 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[9 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[10 + i * 4] = (byte)(pose & 0xFF);
                }
                if (j == 6)
                {
                    data[31] = 0;
                    data[32] = 0;
                    data[33] = 0;
                    data[34] = 0;
                }

                for (int i = 0; i < j; i++)
                {
                    pose = (Int32)(target[i] * 1000.0);

                    data[35 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[36 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[37 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[38 + i * 4] = (byte)(pose & 0xFF);
                }
                if (j == 6)
                {
                    data[59] = 0;
                    data[60] = 0;
                    data[61] = 0;
                    data[62] = 0;
                }

                data[63] = getCRC(data, 63);
                int ret = 0;
                if (sendData(data))
                {
                    byte[] res = await WaitForResponseAsync(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public async Task<byte> CMOVEAsync(byte pointType1, byte pointIndex1, byte pointIndex2, byte spd, byte acc, byte dec)
        {
            byte rn = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[10] { 0x01, 0x02, 0x65, pointType1, pointIndex1, pointIndex2, spd, acc, dec, 0x00 };
                data[9] = getCRC(data, 9);
                int ret = 0;
                if (sendData(data))
                {
                    byte[] res = await WaitForResponseAsync(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public async Task<byte> JAPPROAsync(byte pointType, byte pointIndex, byte spd, byte acc, byte dec, double offset)
        {
            byte rn = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[13];
                int ret = 0;
                Int32 pose = 0;

                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x68;
                data[3] = pointType;
                data[4] = pointIndex;
                data[5] = spd;
                data[6] = acc;
                data[7] = dec;

                pose = (Int32)(offset * 1000.0);

                data[8] = (byte)((pose >> 24) & 0xFF);
                data[9] = (byte)((pose >> 16) & 0xFF);
                data[10] = (byte)((pose >> 8) & 0xFF);
                data[11] = (byte)(pose & 0xFF);

                data[12] = getCRC(data, 12);
                if (sendData(data))
                {
                    byte[] res = await WaitForResponseAsync(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public async Task<byte> LAPPROAsync(byte pointType, byte pointIndex, byte spd, byte acc, byte dec, double offset)
        {
            byte rn = 0;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[13];
                int ret = 0;
                Int32 pose = 0;

                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x69;
                data[3] = pointType;
                data[4] = pointIndex;
                data[5] = spd;
                data[6] = acc;
                data[7] = dec;

                pose = (Int32)(offset * 1000.0);

                data[8] = (byte)((pose >> 24) & 0xFF);
                data[9] = (byte)((pose >> 16) & 0xFF);
                data[10] = (byte)((pose >> 8) & 0xFF);
                data[11] = (byte)(pose & 0xFF);

                data[12] = getCRC(data, 12);
                if (sendData(data))
                {
                    byte[] res = await WaitForResponseAsync(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public async Task<byte> LAPPROAsync(byte pointType, byte spd, byte acc, byte dec, double[] target, double offset)
        {
            byte rn = 0;
            Int32 pose = 0;
            int j = target.Length;
            if (j < 6)
                return 0x22;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[40];
                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x6A;
                data[3] = pointType;
                data[4] = spd;
                data[5] = acc;
                data[6] = dec;

                for (int i = 0; i < j; i++)
                {
                    pose = (Int32)(target[i] * 1000.0);

                    data[7 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[8 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[9 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[10 + i * 4] = (byte)(pose & 0xFF);
                }
                if (j == 6)
                {
                    data[31] = 0;
                    data[32] = 0;
                    data[33] = 0;
                    data[34] = 0;
                }

                pose = (Int32)(offset * 1000.0);

                data[35] = (byte)((pose >> 24) & 0xFF);
                data[36] = (byte)((pose >> 16) & 0xFF);
                data[37] = (byte)((pose >> 8) & 0xFF);
                data[38] = (byte)(pose & 0xFF);

                data[39] = getCRC(data, 39);

                int ret = 0;
                if (sendData(data))
                {
                    byte[] res = await WaitForResponseAsync(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        public async Task<byte> JAPPROAsync(byte pointType, byte spd, byte acc, byte dec, double[] target, double offset)
        {
            byte rn = 0;
            Int32 pose = 0;
            int j = target.Length;
            if (j < 6)
                return 0x22;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[40];
                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x6B;
                data[3] = pointType;
                data[4] = spd;
                data[5] = acc;
                data[6] = dec;

                for (int i = 0; i < j; i++)
                {
                    pose = (Int32)(target[i] * 1000.0);

                    data[7 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[8 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[9 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[10 + i * 4] = (byte)(pose & 0xFF);
                }
                if (j == 6)
                {
                    data[31] = 0;
                    data[32] = 0;
                    data[33] = 0;
                    data[34] = 0;
                }

                pose = (Int32)(offset * 1000.0);

                data[35] = (byte)((pose >> 24) & 0xFF);
                data[36] = (byte)((pose >> 16) & 0xFF);
                data[37] = (byte)((pose >> 8) & 0xFF);
                data[38] = (byte)(pose & 0xFF);

                data[39] = getCRC(data, 39);

                int ret = 0;
                if (sendData(data))
                {
                    byte[] res = await WaitForResponseAsync(startBytes, _motionTimeOut);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0;
                    }
                    else return rn;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }
        /* ---------------------------- RTC commands -------------------------------- */
        public byte RTC_ON()
        {
            return sendCommand(0x51);
        }

        public byte RTC_OFF()
        {
            return sendCommand(0x52);
        }
        public byte setRTC_OffsetData(double x, double y, double z, double rx, double ry, double rz)
        {
            byte ret = 0;
            Int32 pose = 0;
            double[] jt = new double[6] { x, y, z, rx, ry, rz };
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[29];
                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x53;
                for (int i = 0; i < 6; i++)
                {
                    pose = (Int32)(jt[i] * 1000.0);

                    data[3 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[4 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[5 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[6 + i * 4] = (byte)(pose & 0xFF);
                }

                data[28] = getCRC(data, 28);
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, _timeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                            return ret;
                        else
                            return errorCode;
                    }
                    return ret;
                }
                else
                    return 0xFE;
            }
            else
                return 0xFF;
        }

        public byte setRTC_OffsetData(double[] offset)
        {
            if (offset.Length < 6)
                return 0x22;
            return setRTC_OffsetData(offset[0], offset[1], offset[2], offset[3], offset[4], offset[5]);
        }

        public byte RTC_move(byte type, byte period, double[] target)
        {
            byte rn = 0;
            Int32 pose = 0;
            int j = target.Length;
            if (j < 6)
                return 0x22;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[34];
                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x5C;
                data[3] = type;
                data[4] = period;


                for (int i = 0; i < j; i++)
                {
                    pose = (Int32)(target[i] * 1000.0);

                    data[5 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[6 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[7 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[8 + i * 4] = (byte)(pose & 0xFF);
                }
                if (j == 6)
                {
                    data[29] = 0;
                    data[30] = 0;
                    data[31] = 0;
                    data[32] = 0;
                }
                data[33] = getCRC(data, 33);
                int ret = 0;
                if (sendData(data))
                {
                    byte[] res = WaitForResponse(startBytes, 2 * (int)period);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0x10;
                    }
                    else return rn;
                }
                else
                    return 0xFE;

            }
            else
                return 0xFF;
        }

        public async Task<byte> RTC_moveAsync(byte type, byte period, double[] target)
        {
            byte rn = 0;
            Int32 pose = 0;
            int j = target.Length;
            if (j < 6)
                return 0x22;
            if (COM.IsOpen || Client.Connected)
            {
                byte[] data = new byte[34];
                data[0] = 0x01;
                data[1] = 0x02;
                data[2] = 0x5C;
                data[3] = type;
                data[4] = period;


                for (int i = 0; i < j; i++)
                {
                    pose = (Int32)(target[i] * 1000.0);

                    data[5 + i * 4] = (byte)((pose >> 24) & 0xFF);
                    data[6 + i * 4] = (byte)((pose >> 16) & 0xFF);
                    data[7 + i * 4] = (byte)((pose >> 8) & 0xFF);
                    data[8 + i * 4] = (byte)(pose & 0xFF);
                }
                if (j == 6)
                {
                    data[29] = 0;
                    data[30] = 0;
                    data[31] = 0;
                    data[32] = 0;
                }
                data[33] = getCRC(data, 33);
                int ret = 0;
                if (sendData(data))
                {
                    byte[] res = await WaitForResponseAsync(startBytes, 2 * (int)period);
                    rn = ValidateResponse(res);
                    if (rn == 0)
                    {
                        ret = DecodeResponse(res);
                        if (ret == 2)
                            return 0;
                        else if (ret == -1)
                            return errorCode;
                        else
                            return 0x10;
                    }
                    else return rn;
                }
                else
                    return 0xFE;

            }
            else
                return 0xFF;
        }

        /* ---------------------- AS commands ----------------------- */

        public byte executeASCommand(string command)
        {
            byte ret = 0;
            if (COM.IsOpen || Client.Connected)
            {
                if (SendFrame(command, 0x4B))
                {
                    byte[] res = WaitForResponse(startBytes, _motionTimeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                            return ret;
                        else
                            return errorCode;
                    }
                    return ret;
                }
                else
                    return 0xFE;
            }
            else return 0xFF;
        }

        /* ---------------------- AS commands Async ----------------------- */
        public async Task<byte> executeASCommandAsync(string command)
        {
            byte ret = 0;
            if (COM.IsOpen || Client.Connected)
            {
                if (SendFrame(command, 0x4B))
                {
                    byte[] res = await WaitForResponseAsync(startBytes, _motionTimeOut);
                    ret = ValidateResponse(res);
                    if (ret == 0)
                    {
                        if (DecodeResponse(res) >= 0)
                            return ret;
                        else
                            return errorCode;
                    }
                    return ret;
                }
                else
                    return 0xFE;
            }
            else return 0xFF;
        }
    }
}